
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from aionetworking.logging.loggers import BaseLogger, ConnectionLogger


LoggerType = TypeVar('LoggerType', bound='BaseLogger')
ConnectionLoggerType = TypeVar('ConnectionLoggerType', bound='ConnectionLogger')



from logging import Filter, LogRecord

from dataclasses import dataclass
from aionetworking.utils import Expression

from typing import Sequence, Optional


@dataclass
class PeerFilter(Filter):
    peers: Sequence[str]

    def __init__(self, peers: Sequence[str], *args, **kwargs):
        self.peers = peers
        super().__init__(*args, **kwargs)

    def filter(self, record: LogRecord) -> bool:
        address = getattr(record, 'address', None)
        host = getattr(record, 'host', None)
        if self.peers and (address or host):
            return any(s in self.peers for s in (address, host))
        return True


@dataclass
class MessageFilter(Filter):

    expr: Expression

    def __init__(self, expr: Optional[Expression], *args, **kwargs):
        self.expr = expr
        super().__init__(*args, **kwargs)

    def filter(self, record: LogRecord) -> bool:
        msg_obj = getattr(record, 'msg_obj', None)
        if self.expr and msg_obj:
            return self.expr(msg_obj)
        return True



from abc import abstractmethod
import asyncio
from dataclasses import dataclass, field
import logging
from pathlib import Path

from .base import BaseAction
from aionetworking.logging.loggers import get_logger_receiver
from aionetworking import settings
from aionetworking.logging.utils_logging import p
from aionetworking.futures.value_waiters import StatusWaiter
from aionetworking.utils import makedirs, set_task_name
from aionetworking.types.logging import LoggerType

from typing import Any, ClassVar, AnyStr, Protocol
from aionetworking.types.formats import MessageObjectType


@dataclass
class ManagedFile:
    path: Path
    mode: str = 'ab'
    buffering: int = -1
    timeout: int = 10
    retries: int = 3
    retry_interval: int = 2
    logger: LoggerType = field(default_factory=get_logger_receiver)
    _status: StatusWaiter = field(default_factory=StatusWaiter, init=False)
    previous: 'ManagedFile' = field(default=None)
    _queue: asyncio.Queue = field(default_factory=asyncio.Queue, init=False, repr=False, hash=False, compare=False)
    _exception: OSError = field(default=None, init=False)
    _open_files: ClassVar = {}

    @classmethod
    def open(cls, path, *args, **kwargs) -> 'ManagedFile':
        try:
            f = cls._open_files[path]
            if not f.is_closing():
                return f
            kwargs['previous'] = f
        except KeyError:
            pass
        f = cls(path, *args, **kwargs)
        cls._open_files[path] = f
        return f

    @classmethod
    async def close_all(cls, base_path: Path = None) -> None:
        if base_path:
            files = [f for f in cls._open_files.values() if f.is_in(base_path)]
        else:
            files = [f for f in cls._open_files.values()]
        await asyncio.gather(*[f.close() for f in files])

    @classmethod
    def num_files(cls):
        return len(cls._open_files)

    def __post_init__(self):
        self.path.parent.mkdir(parents=True, exist_ok=True)
        self._task = asyncio.create_task(self.manage())
        set_task_name(self._task, f"ManagedFile:{self.path.name}")

    def is_in(self, path) -> bool:
        try:
            self.path.relative_to(path)
            return True
        except ValueError:
            return False

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb): ...

    def is_closing(self):
        return self._status.has_started() and self._status.is_stopping_or_stopped()

    async def wait_closed(self):
        return await self._status.wait_stopped()

    async def wait_has_started(self):
        return await self._status.wait_has_started()

    def _cleanup(self) -> None:
        self._status.set_stopping()
        if self._open_files[self.path] == self:
            del self._open_files[self.path]
        self.logger.debug('Cleanup completed for %s', self.path)
        self._status.set_stopped()

    async def write(self, data: AnyStr) -> None:
        fut = asyncio.Future()
        if self._exception:
            fut.set_exception(self._exception)
        else:
            self._queue.put_nowait((data, fut))
            if self.logger.isEnabledFor(logging.DEBUG):
                qsize = self._queue.qsize()
                self.logger.debug('Message added to queue with future %s', id(fut))
                self.logger.debug('There %s now %s in the write queue %s for %s',
                                  p.plural_verb('is', qsize), p.no('item', qsize), id(self._queue), self.path)
        await fut

    async def close(self) -> None:
        if not self.is_closing():
            await self._status.wait_started()
            self._status.set_stopping()
            self.logger.debug('Closing file %s', self.path)
            if not self._task.done():
                await self.wait_has_started()
                await self.wait_writes_done()
                self._task.cancel()
                try:
                    await self._task
                except asyncio.CancelledError:
                    pass
            else:
                self.logger.debug('File %s already closed', self.path)
            self.logger.debug('Closed file %s', self.path)

    async def wait_writes_done(self) -> None:
        self.logger.debug('Waiting for writes to complete for %s', self.path)
        queue_done_task = asyncio.create_task(self._queue.join())
        done, pending = await asyncio.wait([queue_done_task, self._task], return_when=asyncio.FIRST_COMPLETED)
        for d in done:
            if exc := d.exception():
                self.logger.error(exc)
                await d
        self.logger.debug('Writes completed for %s', self.path)

    def _task_done(self, num: int) -> None:
        for _ in range(0, num):
            self._queue.task_done()
        self.logger.info('Task done set for %s on file %s', p.no('item', num), self.path)

    async def manage_wrapper(self):
        if self.previous:
            await self.previous.wait_closed()
        for i in range(0, self.retries):
            try:
                await self.manage()
                return
            except OSError as e:
                if i == 3:
                    self._exception = e
                    while not self._queue.empty():
                        try:
                            item, fut = self._queue.get_nowait()
                            fut.set_exception(e)
                            self._task_done(1)
                        except asyncio.QueueEmpty:
                            self.logger.info('QueueEmpty error was unexpectedly caught for file %s', self.path)
                await asyncio.sleep(self.retry_interval)

    async def _write_items_from_queue(self, f):
        self.logger.info('Retrieving item from queue for file %s. Timeout: %s', self.path,
                         p.no('second', self.timeout))
        try:
            data, fut = self._queue.get_nowait()
        except asyncio.QueueEmpty:
            data, fut = await asyncio.wait_for(self._queue.get(), timeout=self.timeout)
        futs = [fut]
        try:
            while not self._queue.empty():
                try:
                    item, fut = self._queue.get_nowait()
                    data += item
                    futs.append(fut)
                except asyncio.QueueEmpty:
                    self.logger.warning('QueueEmpty error was unexpectedly caught for file %s', self.path)
            self.logger.info('Retrieved %s from queue. Writing to file %s.', p.no('item',
                                                                                  len(futs)), self.path)
            await f.write(data)
            self.logger.info('%s written to file %s', p.no('byte', len(data)), self.path)
            for fut in futs:
                fut.set_result(True)
                self.logger.debug('Result set on future %s', id(fut))
        except Exception as e:
            for fut in futs:
                fut.set_exception(e)
        finally:
            asyncio.get_running_loop().call_soon(self._task_done, len(futs))

    async def manage(self) -> None:
        self._status.set_starting()
        await makedirs(self.path.parent, exist_ok=True)
        self.logger.info('Opening file %s', self.path)
        async with settings.FILE_OPENER(self.path, mode=self.mode, buffering=self.buffering) as f:
            try:
                self.logger.debug('File %s opened', self.path)
                self._status.set_started()
                while True:
                    try:
                        await self._write_items_from_queue(f)
                    except asyncio.TimeoutError:
                        qsize = self._queue.qsize()
                        if qsize:
                            self.logger.warning(
                                'Get item for file %s timed out out even though there %s %s in the queue id %s',
                                self.path, p.plural_verb('is', qsize), p.no('item', qsize), id(self._queue))
                            await self._write_items_from_queue(f)
                        else:
                            self.logger.info('File %s closing due to timeout on new items to write', self.path)
                            break
            except asyncio.CancelledError as e:
                self.logger.info('File %s closing due to task being cancelled', self.path)
                raise e
            finally:
                self._cleanup()
                qsize = self._queue.qsize()
                if qsize:
                    self.logger.warning(
                        'There %s %s in queue id %s for path %s even after cleanup',
                        p.plural_verb('is', qsize), p.no('item', qsize), id(self._queue), self.path)
                    await self._write_items_from_queue(f)
        self.logger.info('File %s closed', self.path)


def default_data_dir():
    return settings.DATA_DIR


@dataclass
class BaseFileStorage(BaseAction, Protocol):
    base_path: Path = field(default_factory=default_data_dir, metadata={'pickle': True})
    path: str = ''
    attr: str = 'encoded'
    mode: str = 'wb'
    separator: AnyStr = ''

    def __post_init__(self):
        if 'b' in self.mode:
            if isinstance(self.separator, str):
                self.separator = self.separator.encode()
        self._status.is_started()
        settings.ACTION_CONFIG = {
            'base_path': self.base_path,
            'path': self.path,
            'attr': self.attr,
            'mode': self.mode,
            'separator': self.separator
        }

    async def start(self, logger: LoggerType = None) -> None:
        await super().start(logger)
        if not self.quiet:
            print(f'Storing output files in {self.base_path}')

    def _get_full_path(self, msg: MessageObjectType) -> Path:
        return self.base_path / self._get_path(msg)

    def _get_path(self, msg: MessageObjectType) -> Path:
        return Path(self.path.format(msg=msg))

    def _get_data(self, msg: MessageObjectType) -> AnyStr:
        data = getattr(msg, self.attr)
        if self.separator:
            data += self.separator
        return data

    @abstractmethod
    async def _write_to_file(self, path: Path, data: AnyStr): ...

    async def write_one(self, msg: MessageObjectType) -> Path:
        path = self._get_full_path(msg)
        msg.logger.debug('Writing to file %s', path)
        data = self._get_data(msg)
        await self._write_to_file(path, data)
        msg.logger.debug('Data written to file %s', path)
        return path

    async def do_one(self, msg: MessageObjectType) -> Any:
        self._status.set_started()
        if getattr(msg, 'store', True):
            await self.write_one(msg)
        return getattr(msg, 'response', None)


@dataclass
class FileStorage(BaseFileStorage):
    name = 'File Storage'

    async def _write_to_file(self, path: Path, data: AnyStr) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        async with settings.FILE_OPENER(path, self.mode) as f:
            await f.write(data)


@dataclass
class BufferedFileStorage(BaseFileStorage):
    name = 'Buffered File Storage'
    mode: str = 'ab'
    _qsize: int = 0

    close_file_after_inactivity: int = 10
    buffering: int = -1

    async def _write_to_file(self, path: Path, data: AnyStr) -> None:
        async with ManagedFile.open(path, mode=self.mode, buffering=self.buffering,
                                    timeout=self.close_file_after_inactivity, logger=self.logger) as f:
            await f.write(data)

    async def close(self) -> None:
        self._status.set_stopping()
        await ManagedFile.close_all(base_path=self.base_path)
        self._status.set_stopped()



from abc import abstractmethod
from datetime import datetime
from pathlib import Path

from typing import AsyncGenerator, Any, Dict, Sequence, Optional, Protocol
from aionetworking.types.formats import CodecType, MessageObjectType


class MessageObject(Protocol):

    @classmethod
    @abstractmethod
    def _get_codec_kwargs(cls) -> Dict:
        return {}

    @classmethod
    @abstractmethod
    def get_codec(cls, first_buffer_received: Optional[bytes], **kwargs) -> CodecType: ...

    @property
    @abstractmethod
    def full_receiver(self) -> str: ...

    @property
    @abstractmethod
    def receiver(self) -> str: ...

    @property
    @abstractmethod
    def sender(self) -> str: ...

    @property
    @abstractmethod
    def address(self) -> str: ...

    @property
    @abstractmethod
    def full_sender(self) -> str: ...

    @property
    @abstractmethod
    def uid(self) -> Any: ...

    @property
    @abstractmethod
    def request_id(self) -> Any: ...

    @property
    @abstractmethod
    def pformat(self) -> str: ...

    @property
    @abstractmethod
    def timestamp(self) -> datetime: ...

    @abstractmethod
    def filter(self) -> bool: ...


class Codec(Protocol):

    @abstractmethod
    async def decode(self, encoded: bytes, **kwargs) -> AsyncGenerator[Sequence[bytes], None]:
        yield

    @abstractmethod
    def decode_one(self, encoded: bytes, **kwargs) -> Any: ...

    @abstractmethod
    async def encode(self, decoded: Any, **kwargs) -> bytes: ...

    @abstractmethod
    async def decode_buffer(self, encoded: bytes, **kwargs) -> AsyncGenerator[MessageObjectType, None]:
        yield

    @abstractmethod
    async def encode_obj(self, decoded: Any, **kwargs) -> MessageObjectType: ...

    @abstractmethod
    async def from_file(self, file_path: Path, **kwargs) -> AsyncGenerator[MessageObjectType, None]:
        yield

    @abstractmethod
    async def one_from_file(self, file_path: Path, **kwargs) -> MessageObjectType: ...



import asyncio
from dataclasses import dataclass, field
from pathlib import Path
import datetime
from functools import partial
import socket

from .exceptions import MessageFromNotAuthorizedHost

from aionetworking.logging.loggers import get_logger_receiver
from aionetworking.types.logging import LoggerType, ConnectionLoggerType
from aionetworking.types.networking import AFINETContext, AFUNIXContext, NamedPipeContext, BaseContext
from aionetworking.utils import (addr_tuple_to_str, dataclass_getstate, dataclass_setstate, IPNetwork, supernet_of,
                                 hostname_or_ip, set_task_name)
from aionetworking.futures.value_waiters import StatusWaiter

from .connections_manager import connections_manager
from .adaptors import ReceiverAdaptor, SenderAdaptor
from .protocols import (
    ConnectionDataclassProtocol, AdaptorProtocolGetattr, UDPConnectionMixinProtocol, SenderAdaptorGetattr)
from .transports import TransportType, DatagramTransportWrapper
from aionetworking.types.networking import AdaptorType, SenderAdaptorType

from typing import NoReturn, Optional, Tuple, Type, Dict, Any, Sequence, Callable, Awaitable, List, Protocol
from .ssl import check_peercert_expired


AsyncCallable = Callable[[int], Awaitable[None]]


@dataclass
class BaseConnectionProtocol(AdaptorProtocolGetattr, ConnectionDataclassProtocol, Protocol):
    _connected: asyncio.Future = field(default_factory=asyncio.Future, init=False, compare=False)
    _closing: asyncio.Future = field(default_factory=asyncio.Future, init=False, compare=False)
    _status: StatusWaiter = field(default_factory=StatusWaiter, init=False)
    context: BaseContext = field(default_factory=dict, metadata={'pickle': True})
    codec_config: Dict[str, Any] = field(default_factory=dict, metadata={'pickle': True})
    logger: LoggerType = field(default_factory=get_logger_receiver, metadata={'pickle': True})

    def __post_init__(self):
        names = self.parent_name.split(' ')
        self.context['protocol_name'] = f'{names[0]} {names[1]}'

    def __getattr__(self, item):
        if self._adaptor:
            try:
                return getattr(self._adaptor, item)
            except AttributeError:
                raise AttributeError(f"Neither connection nor adaptor have attribute {item}")
        raise AttributeError(f"Connection does not have attribute {item}, and adaptor has not been configured yet")

    def _start_adaptor(self) -> None:
        self._set_adaptor()
        num = connections_manager.add_connection(self)
        self.logger.log_num_connections('opened', num)

    def __getstate__(self):
        return dataclass_getstate(self)

    def __setstate__(self, state):
        dataclass_setstate(self, state)

    @staticmethod
    def get_peername(peer_prefix: str, peer: str, own: str) -> str:
        return f"{peer_prefix}_{own}_{peer}"

    @property
    def peer(self) -> str:
        return self.get_peername(self.peer_prefix, self.context.get('peer'), self.context.get('own'))

    def _set_adaptor(self) -> None:
        kwargs = {
            'context': self.context,
            'dataformat': self.dataformat,
            'preaction': self.preaction,
            'send': self.send,
            'codec_config': self.codec_config,
            'logger': self._get_connection_logger(),
        }
        if self.adaptor_cls.is_receiver:
            self._adaptor = self._get_receiver_adaptor(**kwargs)
        else:
            self._adaptor = self._get_sender_adaptor(**kwargs)

    def _get_receiver_adaptor(self, **kwargs) -> AdaptorType:
        return self.adaptor_cls(action=self.action, **kwargs)

    def _get_sender_adaptor(self, **kwargs) -> SenderAdaptorType:
        return self.adaptor_cls(requester=self.requester, **kwargs)

    def is_child(self, parent_name: str) -> bool:
        return parent_name == self.parent_name

    def _get_connection_logger(self) -> ConnectionLoggerType:
        return self.logger.get_connection_logger(extra=self.context)

    def _delete_connection(self) -> None:
        connections_manager.remove_connection(self)

    async def _close(self, exc: Optional[BaseException]) -> None:
        try:
            if self._adaptor:
                await asyncio.wait_for(self._adaptor.close(exc), timeout=self.timeout)
        finally:
            if self._adaptor:
                num = connections_manager.decrement(self)
                self.logger.log_num_connections('closed', num)
            self._status.set_stopped()

    def finish_connection(self, exc: Optional[BaseException]) -> None:
        if self._adaptor:
            self._adaptor.logger.info('Finishing connection')
            self._delete_connection()
        self._status.set_stopping()
        close_task = asyncio.create_task(self._close(exc))
        set_task_name(close_task, f"Close:{self.peer}")

    def close(self, immediate: bool = False):
        self.finish_connection(None)

    async def wait_closed(self) -> None:
        await self._status.wait_stopped()

    async def wait_connected(self) -> None:
        await self._status.wait_started()

    def is_closing(self) -> bool:
        return self._status.is_stopping_or_stopped()

    def is_connected(self) -> bool:
        return self._status.is_started()

    def data_received(self, data: bytes) -> None:
        self.last_msg = datetime.datetime.now()
        self._adaptor.on_data_received(data, timestamp=self.last_msg)


@dataclass
class NetworkConnectionProtocol(BaseConnectionProtocol, Protocol):
    transport: asyncio.BaseTransport = field(default=None, init=False)
    pause_reading_on_buffer_size: int = None
    allowed_senders: Sequence[IPNetwork] = field(default_factory=tuple)
    hostname_lookup: bool = False
    connection_lost_tasks: List[AsyncCallable] = field(default_factory=list)
    check_peer_cert_expiry: int = 7
    _unprocessed_data: int = field(default=0, init=False, repr=False)

    def _raise_message_from_not_authorized_host(self, host: str) -> NoReturn:
        msg = f"Received message from unauthorized host {host}"
        self.logger.error(msg)
        raise MessageFromNotAuthorizedHost(msg)

    def _sender_valid(self, peer_ip, peer_hostname):
        if self.allowed_senders:
            return supernet_of(peer_ip, peer_hostname, self.allowed_senders)
        return True

    def _check_peer(self) -> None:
        if not self._sender_valid(self.context['address'], self.context['host']):
            self._raise_message_from_not_authorized_host(self.context['host'])

    def close(self, immediate: bool = False):
        if not self.transport.is_closing():
            self.transport.close()

    def log_context(self):
        self.logger.info(self.context)

    def initialize_connection(self, transport: TransportType, **extra_context) -> bool:
        self._status.set_starting()
        self._update_context(transport)
        self.context.update(extra_context)
        try:
            if self.context.get('host'):
                self._check_peer()
            self.last_msg = datetime.datetime.now()
            self._start_adaptor()
            self._status.set_started()
            return True
        except MessageFromNotAuthorizedHost:
            self.close(immediate=True)
            return False

    def add_connection_lost_task(self, async_function: AsyncCallable):
        self.connection_lost_tasks.append(async_function)

    def run_connection_lost_tasks(self) -> None:
        if self.connection_lost_tasks:
            [asyncio.create_task(t()) for t in self.connection_lost_tasks]

    def eof_received(self) -> bool:
        return False

    def _update_context(self, transport: TransportType) -> None:
        _socket = transport.get_extra_info('socket', None)
        if _socket and hasattr(socket, 'AF_UNIX') and _socket.family == socket.AF_UNIX:
            # AF_UNIX server transport
            sockname: str = transport.get_extra_info('sockname') or transport.get_extra_info('peername')
            fd: int = _socket.fileno()
            self.context: AFUNIXContext = {
                'protocol_name': self.context['protocol_name'],
                'address': Path(sockname).name,
                'peer': str(fd) if self.adaptor_cls.is_receiver else sockname,
                'own': sockname if self.adaptor_cls.is_receiver else str(fd),
                'alias': f'{sockname}.{fd}',
                'server': sockname,
                'client': str(fd),
                'fd': fd,
            }
        elif transport.get_extra_info('pipe', None):
            # Windows Named Pipe Transport
            addr: str = transport.get_extra_info('addr')
            handle: int = transport.get_extra_info('pipe').handle
            peer: str = str(handle) if self.adaptor_cls.is_receiver else addr
            self.context: NamedPipeContext = {
                'protocol_name': self.context['protocol_name'],
                'address': Path(addr).name,
                'server': addr,
                'handle': handle,
                'client': str(handle),
                'own': addr if self.adaptor_cls.is_receiver else str(handle),
                'peer': peer,
                'alias': f'{addr}.{handle}'
            }
        else:
            # INET/INET6 transport
            peer: Tuple[str, int] = transport.get_extra_info('peername')[0:2]
            sockname: Tuple[str, int] = transport.get_extra_info('sockname')[0:2]
            peer_str = addr_tuple_to_str(peer)
            sock_str = addr_tuple_to_str(sockname)
            if self.hostname_lookup:
                host = hostname_or_ip(peer[0])
            else:
                host = peer[0]
            self.context: AFINETContext = {
                'protocol_name': self.context['protocol_name'],
                'peer': peer_str,
                'own': sock_str,
                'address': peer[0],
                'port': peer[1],
                'host': host,
                'alias': '',
                'server': sock_str if self.adaptor_cls.is_receiver else peer_str,
                'client': peer_str if self.adaptor_cls.is_receiver else sock_str
            }
            if self.context['host'] not in self.context['address']:
                self.context['alias'] = f"{self.context['host']}({self.context['peer']})"
            else:
                self.context['alias'] = self.context['host']
        self.log_context()
        cipher = transport.get_extra_info('cipher', default=None)
        if cipher:
            self.logger.info(f"Cipher: {cipher}', 'Compression': {transport.get_extra_info('compression')}")
            peercert = transport.get_extra_info('peercert')
            if peercert:
                self.logger.debug(peercert)
                if self.check_peer_cert_expiry:
                    expiry_time, cert_expiry_in_days = check_peercert_expired(peercert, self.check_peer_cert_expiry)
                    if cert_expiry_in_days:
                        self.logger.warn_on_cert_expiry(f'Peer @ {self.context["host"]}', cert_expiry_in_days, expiry_time)

    def send(self, msg: bytes) -> None:
        self.transport.write(msg)
        self.last_msg = datetime.datetime.now()


@dataclass
class BaseStreamConnection(NetworkConnectionProtocol, Protocol):
    transport: asyncio.Transport = field(default=None, init=False, repr=False, compare=False)

    def connection_made(self, transport: asyncio.Transport) -> None:
        self.transport = transport
        self.initialize_connection(transport)

    def _close_transport(self, task: asyncio.Future):
        self.transport.close()

    def close(self, immediate: bool = False):
        if not self.transport.is_closing():
            if immediate:
                self.transport.abort()
            elif self._adaptor:
                task = asyncio.create_task(self.wait_current_tasks())
                task.add_done_callback(self._close_transport)
            else:
                self.transport.close()

    def connection_lost(self, exc: Optional[BaseException]) -> None:
        self.close()
        self.run_connection_lost_tasks()
        self.finish_connection(exc)

    def _resume_reading(self, datalen: int, fut: asyncio.Future):
        self._unprocessed_data -= datalen
        if (self.pause_reading_on_buffer_size is not None and not self.transport.is_reading() and
                not self.transport.is_closing()):
            if self.pause_reading_on_buffer_size >= self._unprocessed_data:
                self.transport.resume_reading()
                self._adaptor.logger.info('Reading resumed')
        fut.result()

    def data_received(self, data: bytes) -> None:
        self.last_msg = datetime.datetime.now()
        self._unprocessed_data += len(data)
        if self.pause_reading_on_buffer_size is not None:
            if self.pause_reading_on_buffer_size <= len(data):
                self.transport.pause_reading()
                self.logger.info('Reading Paused')
        task = self._adaptor.on_data_received(data, timestamp=self.last_msg)
        task.add_done_callback(partial(self._resume_reading, len(data)))


@dataclass
class BaseUDPConnection(NetworkConnectionProtocol, UDPConnectionMixinProtocol):
    _peer: Tuple[str, int] = field(default=None, init=False)
    transport: DatagramTransportWrapper = field(default=None, init=False, repr=False, compare=False)

    def connection_made(self, transport: DatagramTransportWrapper) -> bool:
        self.transport = transport
        return self.initialize_connection(transport)

    def connection_lost(self, exc: Optional[BaseException]) -> None:
        self.run_connection_lost_tasks()
        self.finish_connection(exc)


@dataclass
class TCPServerConnection(BaseStreamConnection):
    name = 'TCP Server'
    adaptor_cls: Type[AdaptorType] = ReceiverAdaptor


@dataclass
class TCPClientConnection(BaseStreamConnection, SenderAdaptorGetattr, asyncio.Protocol):
    name = 'TCP Client'
    adaptor_cls: Type[AdaptorType] = SenderAdaptor


@dataclass
class UDPServerConnection(BaseUDPConnection):
    name = 'UDP Server'
    adaptor_cls: Type[AdaptorType] = ReceiverAdaptor


@dataclass
class UDPClientConnection(BaseUDPConnection, SenderAdaptorGetattr):
    name = 'UDP Client'
    adaptor_cls: Type[AdaptorType] = SenderAdaptor




from abc import abstractmethod
from dataclasses import dataclass, field

from aionetworking.logging.loggers import Logger, get_connection_logger_sender
from typing import Protocol


@dataclass
class RequesterProtocol(Protocol):
    name = 'sender'
    methods = ()
    notification_methods = ()

    logger: Logger = field(default_factory=get_connection_logger_sender, compare=False)

    @abstractmethod
    async def start(self): ...

    @abstractmethod
    async def close(self): ...



from setuptools import setup

import platform
from pathlib import Path
from pkg_resources import parse_requirements

from typing import List, Union


def get_requirements(filename: Union[str, Path]) -> List[str]:
    text = Path(filename).read_text()
    return [str(requirement) for requirement in parse_requirements(text)]


def linux_package_is_installed(*package_names: str) -> bool:
    try:
        import apt
        cache = apt.Cache()
        return all(cache.get(package) and cache[package].is_installed for package in package_names)
    except ImportError:
        try:
            import yum
            yb = yum.YumBase()
            return all(yb.rpmdb.searchNevra(name=package) for package in package_names)
        except ImportError:
            return False


def libsystemd_is_installed() -> bool:
    if platform.system() == 'Linux':
        return linux_package_is_installed('libsystemd-dev')
    return False


readme = Path('README.md').read_text()


required = get_requirements('requirements.txt')
if libsystemd_is_installed():
    required += get_requirements('requirements_systemd.txt')
optional = get_requirements('requirements_optional.txt')
sftp_requirements = get_requirements('requirements_sftp.txt')


# https://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
main_ns = {}
with open('aionetworking/__version__.py') as f:
    exec(f.read(), main_ns)


setup(
    name='aionetworking',
    version=main_ns['version'],
    packages=['aionetworking', 'aionetworking.conf', 'aionetworking.types', 'aionetworking.actions',
              'aionetworking.formats', 'aionetworking.formats.contrib', 'aionetworking.futures',
              'aionetworking.logging', 'aionetworking.senders', 'aionetworking.receivers', 'aionetworking.networking',
              'aionetworking.requesters'],
    scripts=['scripts/generate_ssh_host_key.py'],
    url='https://github.com/primal100/aionetworking',
    license="MIT License",
    author='Unknown',
    description='Various utilities for asyncio networking',
    long_description=readme,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Framework :: AsyncIO',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    setup_requires=['wheel'],
    install_requires=required,
    extras_require={
        'sftp': sftp_requirements,
        'optional': optional,
        'all': sftp_requirements + optional
    }
)



import asyncio
import yaml
import os
from dataclasses import dataclass, field

from logging.config import dictConfig
from aionetworking.actions.yaml_constructors import load_file_storage, load_buffered_file_storage, load_echo_action, \
    load_empty_action
from aionetworking.compatibility_os import loop_on_close_signal, loop_on_user1_signal, send_status, \
    send_ready, send_reloading
from aionetworking.conf.yaml_constructors import load_logger, load_receiver_logger, load_sender_logger
from aionetworking.formats.contrib.yaml_constructors import load_json, load_pickle
from aionetworking.logging.loggers import get_logger_receiver
from aionetworking.networking.yaml_constructors import (load_server_side_ssl, load_client_side_ssl,
                                                        load_stream_server_protocol_factory,
                                                        load_datagram_server_protocol_factory,
                                                        load_stream_client_protocol_factory,
                                                        load_datagram_client_protocol_factory)
from aionetworking.receivers.yaml_constructors import load_tcp_server, load_udp_server, load_pipe_server
from aionetworking.requesters.yaml_constructors import load_echo_requester
from aionetworking.types.logging import LoggerType
from aionetworking.types.receivers import ReceiverType
from aionetworking.types.senders import SenderType
from aionetworking.senders.yaml_constructors import load_tcp_client, load_udp_client, load_pipe_client
from .yaml_constructors import load_ip_network, load_path, load_default_ports, load_env_variable
from aionetworking import settings

from pathlib import Path
from typing import Union, Dict, TextIO


def get_paths(app_home: Union[str, Path] = None, volatile_home: Union[str, Path] = None,
              **kwargs) -> Dict[str, Path]:
    app_home = app_home or settings.APP_HOME
    volatile_home = volatile_home or app_home
    paths = {'temp': settings.TEMPDIR,
            'home': app_home,
            'conf': app_home / 'conf',
            'data': volatile_home / 'data',
            'logs': volatile_home / 'logs',
            'misc': volatile_home / 'misc',
            'stats': volatile_home / 'stats',
            'ssl': app_home / 'ssl'}
    paths.update(**kwargs)
    return paths


def load_minimal_tags() -> None:
    load_env_variable()
    load_default_ports()
    load_logger()
    load_receiver_logger()
    load_sender_logger()
    load_tcp_server()
    load_tcp_client()
    load_udp_server()
    load_udp_client()
    load_pipe_server()
    load_pipe_client()
    load_server_side_ssl()
    load_client_side_ssl()
    load_stream_server_protocol_factory()
    load_datagram_server_protocol_factory()
    load_stream_client_protocol_factory()
    load_datagram_client_protocol_factory()
    load_json()
    load_pickle()
    load_ip_network()
    load_echo_action()
    load_empty_action()
    load_buffered_file_storage()
    load_file_storage()
    load_echo_requester()


def load_all_tags():
    load_minimal_tags()
    from aionetworking.networking.yaml_constructors_sftp import (load_sftp_server_protocol_factory,
                                                                 load_sftp_client_protocol_factory)
    from aionetworking.receivers.yaml_constructors_sftp import load_sftp_server
    from aionetworking.senders.yaml_constructors_sftp import load_sftp_client
    load_sftp_server_protocol_factory()
    load_sftp_client_protocol_factory()
    load_sftp_server()
    load_sftp_client()


def configure_logging(path: Path):
    with path.open('rt') as f:
        config = yaml.safe_load(f.read())
        for handler in config['handlers'].values():
            filename = handler.get('filename')
            if filename:
                filename.parent.mkdir(parents=True, exist_ok=True)
        dictConfig(config)


def node_from_file(path: Union[str, Path], paths: Dict[str, Union[str, Path]] = None) -> \
        Union[ReceiverType, SenderType]:
    stream = open(path, 'r')
    parent = Path(path).parent
    return node_from_config(stream, paths=paths, parent=parent)


def node_from_config(conf: TextIO, paths: Dict[str, Union[str, Path]] = None, parent: Path = None) -> \
        Union[ReceiverType, SenderType]:
    paths = paths or get_paths()
    paths['this'] = parent
    load_path(paths)
    configs = list(yaml.safe_load_all(conf))
    node = configs[0]
    if len(configs) > 1:
        misc_config = configs[1]
        log_config_file = misc_config.pop('log_config_file')
        if not (node_name := misc_config.pop('node_name', None)):
            node_name = node.full_name.replace(' ', '_').replace(':', '_').lower()
        paths['node'] = node_name
        paths['name'] = node.name.replace(' ', '_').lower()
        if '_' in paths['name']:
            paths['type'] = paths['name'].split('_')[1]
        else:
            paths['type'] = paths['name']
        paths['host'] = getattr(node, 'host', '').replace(':', '').lower()
        paths['port'] = getattr(node, 'port', '')
        paths['pipe_path'] = getattr(node, 'path', '')
        paths['pid'] = str(os.getpid())
        configure_logging(log_config_file)
        settings.APP_CONFIG.update(misc_config)
    return node


def node_from_config_file(conf_path: Union[Path, str], paths: Dict[str, Union[str, Path]] = None) -> Union[ReceiverType, SenderType]:
    f = open(str(conf_path), 'r')
    parent_path = Path(conf_path).parent
    return node_from_config(f, paths=paths, parent=parent_path)


def server_from_config_file(conf_path: Union[Path, str], paths: Dict[str, Union[str, Path]] = None) -> ReceiverType:
    return node_from_config_file(conf_path, paths=paths)


def client_from_config_file(conf_path: Union[Path, str], paths: Dict[str, Union[str, Path]] = None) -> SenderType:
    return node_from_config_file(conf_path, paths=paths)


@dataclass
class SignalServerManager:
    conf_path: Union[Path, str]
    server: ReceiverType = field(init=False)
    notify_pid: int = None
    paths: Dict[str, Union[str, Path]] = None
    logger: LoggerType = field(default_factory=get_logger_receiver)
    _last_modified_time: float = field(init=False, default=None)

    def __post_init__(self):
        self._last_modified_time = self.modified_time
        self._stop_event = asyncio.Event()
        self._restart_event = asyncio.Event()
        self._restart_event.set()
        self.server = self.get_server()
        loop_on_close_signal(self.close, self.logger)
        loop_on_user1_signal(self.check_reload, self.logger)

    @property
    def modified_time(self) -> float:
        return os.stat(self.conf_path).st_mtime

    def close(self) -> None:
        self._stop_event.set()

    def check_reload(self) -> None:
        if os.path.exists(self.conf_path):
            last_modified = self.modified_time
            if last_modified > self._last_modified_time:
                send_reloading()
                self._last_modified_time = last_modified
                self.logger.info('Restarting server')
                send_status('Restarting server')
                self._restart_event.set()
            else:
                send_ready()
        else:
            self._stop_event.set()

    def server_is_started(self):
        return self.server.is_started

    async def wait_server_stopped(self):
        await self.server.wait_stopped()

    async def wait_server_started(self):
        await self.server.wait_started()

    def get_server(self) -> ReceiverType:
        return server_from_config_file(self.conf_path, self.paths)

    async def serve_until_stopped(self) -> None:
        while self._restart_event.is_set():
            self._restart_event.clear()
            await self.server.serve_until_close_signal(stop_event=self._stop_event, restart_event=self._restart_event,
                                                       notify_pid=self.notify_pid)
            if self._restart_event.is_set():
                self.server = self.get_server()



from dataclasses import dataclass
from .base import BaseMessageObject
from collections import namedtuple
from pathlib import Path
from .contrib.pickle import PickleCodec
from typing import AsyncGenerator


recorded_packet = namedtuple("recorded_packet", ["sent_by_server", "timestamp", "sender", "data"])


@dataclass
class BufferCodec(PickleCodec):
    log_msgs = False

    async def decode(self, encoded: bytes, **kwargs) -> recorded_packet:
        async for encoded, decoded in super().decode(encoded, **kwargs):
            yield encoded, recorded_packet(*decoded)

    async def encode(self, decoded: bytes, system_timestamp=None, **kwargs) -> bytes:
        if self.context:
            sender = self.context.get('address')
        else:
            sender = None
        packet_data = (
            False,
            system_timestamp,
            sender,
            decoded
        )
        return await super().encode(packet_data, **kwargs)


@dataclass
class BufferObject(BaseMessageObject):
    name = 'Buffer'
    codec_cls = BufferCodec


def get_recording_codec() -> PickleCodec:
    return BufferCodec(BufferObject)


async def get_recording(data: bytes) -> AsyncGenerator[recorded_packet, None]:
    codec = get_recording_codec()
    async for item in codec.decode_buffer(data):
        yield item.decoded


async def get_recording_from_file(path: Path) -> AsyncGenerator[recorded_packet, None]:
    codec = get_recording_codec()
    async for item in codec.from_file(path):
        yield item.decoded





from abc import ABC, abstractmethod
import binascii
import datetime
import logging
from dataclasses import dataclass, field

from aionetworking.utils import dataclass_getstate, dataclass_setstate, get_current_task_name
from aionetworking.utils import SystemInfo, supports_system_info
from aionetworking.logging.utils_logging import LoggingDatetime, LoggingTimeDelta, BytesSize, MsgsCount, BytesSizeRate, p
from aionetworking.futures.schedulers import TaskScheduler

from typing import Type, Optional, Dict, Generator, Any, Union
from aionetworking.types.formats import MessageObjectType


class BaseLogger(logging.LoggerAdapter, ABC):

    def update_extra(self, **kwargs):
        self.extra.update(**kwargs)

    def manage_error(self, exc: BaseException) -> None:
        if exc:
            self.error(exc, exc_info=exc)

    def manage_critical_error(self, exc: BaseException) -> None:
        if exc:
            self.critical(exc, exc_info=exc)

    @abstractmethod
    def _get_connection_logger_cls(self) -> Type: ...

    @abstractmethod
    def get_connection_logger(self, name: str = 'connection', **kwargs): ...

    @abstractmethod
    def get_child(self, name: str, cls: Type = None, **kwargs) -> 'BaseLogger': ...

    @abstractmethod
    def get_sibling(self, name: str, **kwargs) -> 'BaseLogger': ...

    @abstractmethod
    def _get_logger(self, name: str = '', cls: Type['BaseLogger'] = None, extra: Dict[str, Any] = None, **kwargs) -> Any: ...

    @abstractmethod
    def warn_on_cert_expiry(self, cert_name: str, num_days: int, expiry_time: datetime.datetime): ...

    async def wait_closed(self): ...


@dataclass
class Logger(BaseLogger):
    name: str
    datefmt: str = '%Y-%m-%d %H:%M:%S.%f'
    extra: dict = None
    stats_interval: Union[float, int] = 60
    stats_fixed_start_time: bool = True
    is_closing: bool = field(default=False, init=False)

    def __init__(self, name: str, datefmt: str = '%Y-%m-%d %H:%M:%S.%f', extra: Dict = None,
                 stats_interval: Optional[Union[int, float]] = 0, stats_fixed_start_time: bool = True):
        self.logger_name = name
        self.datefmt = datefmt
        self.stats_interval = stats_interval
        self.stats_fixed_start_time = stats_fixed_start_time
        logger = logging.getLogger(name)
        super().__init__(logger, extra or {})

    def __getstate__(self):
        state = dataclass_getstate(self)
        if self.is_closing:
            state['is_closing'] = self.is_closing
        return state

    def __setstate__(self, state):
        self.is_closing = state.pop('is_closing', self.is_closing)
        dataclass_setstate(self, state)

    def process(self, msg, kwargs):
        msg, kwargs = super().process(msg, kwargs)
        kwargs['extra'] = kwargs['extra'].copy()
        kwargs['extra']['taskname'] = get_current_task_name()
        if supports_system_info:
            kwargs['extra']['system'] = SystemInfo()
        kwargs['extra'].update(kwargs.pop('detail', {}))
        return msg, kwargs

    def _get_connection_logger_cls(self) -> Type[BaseLogger]:
        if self._get_child_logger('stats').isEnabledFor(logging.INFO):
            return ConnectionLoggerStats
        return ConnectionLogger

    def _get_child_logger(self, name):
        child_name = f"{self.logger_name}.{name}"
        return logging.getLogger(child_name)

    def get_connection_logger(self, name: str = 'connection', **kwargs) -> Any:
        connection_logger_cls = self._get_connection_logger_cls()
        return self.get_child(name, cls=connection_logger_cls, stats_interval=self.stats_interval,
                              stats_fixed_start_time=self.stats_fixed_start_time, **kwargs)

    def get_child(self, name: str, cls: Type = None, **kwargs) -> Any:
        logger_name = f"{self.logger_name}.{name}"
        return self._get_logger(name=logger_name, cls=cls, **kwargs)

    def get_sibling(self, name: str, *args, **kwargs) -> Any:
        name = f'{self.logger.parent.name}.{name}'
        return self._get_logger(*args, name=name, **kwargs)

    def _get_logger(self, name: str = '', cls: Type[BaseLogger] = None, extra: Dict[str, Any] = None, **kwargs) -> Any:
        cls = cls or self.__class__
        extra = extra or {}
        extra.update(self.extra)
        return cls(name, extra=extra, **kwargs)

    def log_num_connections(self, action: str, num_connections: int):
        if self.isEnabledFor(logging.DEBUG):
            self.log(logging.DEBUG, 'Connection %s. There %s now %s.', action,
                     p.plural_verb('is', p.num(num_connections)),
                     p.no('active connection'))

    def warn_on_cert_expiry(self, cert_name: str, num_days: int, expiry_time: datetime.datetime):
        self.logger.warning('%s ssl cert will expire in less than %s, on %s', cert_name, p.no('day', num_days + 1),
                            expiry_time)

    def _set_closing(self) -> None:
        self.is_closing = True


default_extra: Dict[str, Any] = {
            'endpoint': None,
            'protocol_name': None,
            'server': None,
            'client': None,
            'alias': None,
            'peer': None,
        }


@dataclass
class ConnectionLogger(Logger):

    def __init__(self, *args, extra: Dict[str, Any] = None, **kwargs):
        extra = extra or default_extra
        super().__init__(*args, extra=extra, **kwargs)
        self._raw_received_logger = self.get_sibling('raw_received', cls=Logger)
        self._raw_sent_logger = self.get_sibling('raw_sent', cls=Logger)
        self._msg_received_logger = self.get_sibling('msg_received', cls=Logger)
        self._msg_sent_logger = self.get_sibling('msg_sent', cls=Logger)

    def new_msg_logger(self, msg_obj: MessageObjectType):
        return self.get_sibling("msg", cls=Logger, extra={'msg_obj': msg_obj})

    @property
    def connection_type(self) -> str:
        return self.extra['protocol_name']

    @property
    def client(self) -> str:
        return self.extra['client']

    @property
    def server(self) -> str:
        return self.extra['server']

    @property
    def peer(self) -> str:
        return self.extra['peer']

    @staticmethod
    def _convert_raw_to_hex(data: bytes):
        try:
            return data.decode('utf-8')
        except UnicodeDecodeError:
            return binascii.hexlify(data).decode('utf-8')

    def _raw_received(self, data: bytes, level: int, *args, **kwargs) -> None:
        if self._raw_received_logger.isEnabledFor(level):
            msg = self._convert_raw_to_hex(data)
            self._raw_received_logger.log(level, msg, *args, **kwargs)

    def _raw_sent(self, data: bytes, level: int, *args, **kwargs) -> None:
        if self._raw_sent_logger.isEnabledFor(level):
            msg = self._convert_raw_to_hex(data)
            self._raw_sent_logger.log(level, msg, *args, **kwargs)

    def manage_decode_error(self, buffer: bytes, exc: BaseException):
        self._raw_received(buffer, logging.ERROR)
        self.error('Failed to decode message')
        self.manage_error(exc)

    def _msg_received(self, msg_obj: MessageObjectType, *args, msg: str = '', extra: Dict[str, Any] = None, **kwargs) -> None:
        extra = extra or {}
        extra.update({'msg_obj': msg_obj, 'direction': 'RECEIVED'})
        self._msg_received_logger.debug(msg, *args, detail=extra, **kwargs)

    def _msg_sent(self, msg_obj: MessageObjectType, level: int, *args, msg: str = '', **kwargs) -> None:
        self._msg_sent_logger.log(level, msg, *args, detail={'msg_obj': msg_obj, 'direction': 'SENT'}, **kwargs)

    def on_msg_decoded(self, msg_obj: MessageObjectType, **kwargs) -> None:
        self._msg_received(msg_obj, **kwargs)

    def new_connection(self) -> None:
        self.info('New %s connection from %s to %s', self.connection_type, self.client, self.server)
        self.info(self.extra)

    def on_buffer_received(self, data: bytes) -> None:
        self.info("Received buffer containing %s bytes", len(data))

    def on_buffer_decoded(self, data: bytes, num: int, source: str = 'buffer') -> None:
        self._raw_received(data, logging.DEBUG)
        self.info("Decoded %s in %s", p.no('message', num), source)

    def on_encode_failed(self, msg_obj: MessageObjectType, exc: BaseException):
        self._msg_sent(msg_obj, logging.ERROR)
        self.error('Failed to encode message %s', msg_obj)
        self.manage_error(exc)

    def on_sending_decoded_msg(self, msg_obj: MessageObjectType) -> None:
        self._msg_sent(msg_obj, logging.DEBUG)

    def on_sending_encoded_msg(self, data: bytes) -> None:
        self.debug("Sending message")
        self._raw_sent(data, logging.DEBUG)

    def on_msg_sent(self, data: bytes) -> None:
        self.debug('Message sent')

    def on_msg_processed(self, msg: MessageObjectType) -> None:
        self.debug('Finished processing message %s', msg.uid)

    def on_msg_filtered(self, msg: MessageObjectType) -> None:
        self.debug('Filtered msg %s', msg.uid)

    def on_msg_failed(self, msg: MessageObjectType, exc: BaseException) -> None:
        self.error('Failed to process msg %s', getattr(msg, 'uid', None))
        self.manage_error(exc)

    def connection_finished(self, exc: Optional[BaseException] = None) -> None:
        self.manage_error(exc)
        self.info('%s connection from %s to %s has been closed', self.connection_type, self.client, self.server)


@dataclass
class StatsTracker:
    datefmt: str = '%Y-%m-%d %H:%M:%S.%f'
    start: LoggingDatetime = field(default=None, init=False)
    end: LoggingDatetime = field(default=None, init=False)
    sent: BytesSize = field(default_factory=BytesSize, init=False)
    received: BytesSize = field(default_factory=BytesSize, init=False)
    processed: BytesSize = field(default_factory=BytesSize, init=False)
    filtered: BytesSize = field(default_factory=BytesSize, init=False)
    failed: BytesSize = field(default_factory=BytesSize, init=False)
    largest_buffer: BytesSize = field(default_factory=BytesSize, init=False)
    msgs: MsgsCount = field(default_factory=MsgsCount, init=False)

    attrs = ('start', 'end', 'msgs', 'sent', 'received', 'processed', 'filtered', 'failed', 'largest_buffer',
             'send_rate', 'processing_rate', 'receive_rate', 'interval', 'average_buffer_size', 'average_sent',
             'msgs_per_buffer', 'not_decoded', 'not_decoded_rate', 'total_done')

    def __post_init__(self):
        self.start = LoggingDatetime(datefmt=self.datefmt)

    @property
    def total_done(self) -> int:
        return self.processed + self.failed + self.filtered

    @property
    def processing_rate(self) -> BytesSizeRate:
        return self.processed / (self.msgs.processing_time or 1)

    @property
    def receive_rate(self) -> BytesSizeRate:
        return self.received / (self.msgs.receive_interval or 1)

    @property
    def send_rate(self) -> float:
        return self.sent / (self.msgs.send_interval or 1)

    @property
    def msgs_per_buffer(self) -> float:
        return self.msgs.total_done / (self.msgs.received or 1)

    @property
    def interval(self) -> LoggingTimeDelta:
        return LoggingTimeDelta(self.start, self.end)

    @property
    def average_buffer_size(self) -> BytesSizeRate:
        return self.received / (self.msgs.received or 1)

    @property
    def average_sent(self) -> float:
        return self.sent / (self.msgs.sent or 1)

    @property
    def not_decoded(self) -> int:
        return self.received - self.total_done

    @property
    def not_decoded_rate(self) -> float:
        return self.not_decoded / (self.received or 1)

    def __iter__(self) -> Generator[Any, None, None]:
        yield from self.attrs

    def __getitem__(self, item: Any) -> Any:
        return getattr(self, item)

    def on_buffer_received(self, data: bytes) -> None:
        if not self.msgs.first_received:
            self.msgs.first_received = LoggingDatetime(self.datefmt)
        self.msgs.last_received = LoggingDatetime(self.datefmt)
        self.msgs.received += 1
        size = len(data)
        self.received += size
        if size > self.largest_buffer:
            self.largest_buffer = BytesSize(size)

    def on_msg_processed(self, data: bytes) -> None:
        self.msgs.last_processed = LoggingDatetime(self.datefmt)
        self.msgs.processed += 1
        self.processed += len(data)

    def on_msg_filtered(self, data: bytes) -> None:
        self.msgs.filtered += 1
        self.filtered += len(data)

    def on_msg_failed(self, data: bytes) -> None:
        self.msgs.failed += 1
        self.failed += len(data)

    def on_msg_sent(self, msg: bytes) -> None:
        self.msgs.last_sent = LoggingDatetime(self.datefmt)
        if not self.msgs.first_sent:
            self.msgs.first_sent = self.msgs.last_sent
        self.msgs.sent += 1
        self.sent += len(msg)

    def end_interval(self) -> None:
        self.end = LoggingDatetime(self.datefmt)


@dataclass
class StatsLogger(Logger):
    _first = True
    _stats: StatsTracker = field(default=None, init=False, compare=False)
    _scheduler: TaskScheduler = field(init=False, default_factory=TaskScheduler, compare=False)
    stats_cls = StatsTracker

    def __init__(self, logger_name: str, extra: dict, *args, **kwargs):
        self._logged_last = False
        self._scheduler = TaskScheduler()
        super().__init__(logger_name, *args, extra=extra, **kwargs)
        self._stats = self.stats_cls(datefmt=self.datefmt)
        if self.stats_interval:
            self._scheduler.call_cb_periodic(self.stats_interval, self.periodic_log,
                                             fixed_start_time=self.stats_fixed_start_time)

    def process(self, msg, kwargs):
        self._stats.end_interval()
        msg, kwargs = super().process(msg, kwargs)
        kwargs['extra'].update({k: self._stats[k] for k in self._stats})
        return msg, kwargs

    def reset(self):
        self._stats = self.stats_cls(datefmt=self.datefmt)

    def stats(self, tag: str) -> None:
        self._first = False
        self.info(tag)
        self.reset()

    def periodic_log(self) -> None:
        self.stats("INTERVAL")

    def _log_end(self) -> None:
        self._stats.end_interval()
        tag = 'ALL' if self._first else 'END'
        self.stats(tag)
        self._logged_last = True

    def connection_finished(self):
        self._set_closing()
        self._scheduler.close_nowait()
        self._log_end()

    async def wait_closed(self):
        await self._scheduler.close()

    def on_msg_filtered(self, data: bytes):
        self._stats.on_msg_filtered(data)

    def on_msg_processed(self, data: bytes):
        self._stats.on_msg_processed(data)

    def on_msg_failed(self, data: bytes):
        self._stats.on_msg_failed(data)

    def __getattr__(self, item):
        if self._stats:
            return getattr(self._stats, item)


@dataclass
class ConnectionLoggerStats(ConnectionLogger):
    stats_cls = StatsLogger

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._stats_logger = self._get_stats_logger()

    def _get_stats_logger(self) -> StatsLogger:
        return self.get_sibling('stats', cls=self.stats_cls, stats_interval=self.stats_interval,
                                stats_fixed_start_time=self.stats_fixed_start_time)

    def on_buffer_received(self, msg: bytes) -> None:
        super().on_buffer_received(msg)
        self._stats_logger.on_buffer_received(msg)

    def on_msg_processed(self, msg: MessageObjectType) -> None:
        super().on_msg_processed(msg)
        self._stats_logger.on_msg_processed(msg.encoded)

    def on_msg_filtered(self, msg: MessageObjectType) -> None:
        super().on_msg_processed(msg)
        self._stats_logger.on_msg_filtered(msg.encoded)

    def on_msg_failed(self, msg: MessageObjectType, exc: BaseException) -> None:
        super().on_msg_failed(msg, exc)
        self._stats_logger.on_msg_failed(msg.encoded)

    def on_msg_sent(self, msg: bytes) -> None:
        super().on_msg_sent(msg)
        self._stats_logger.on_msg_sent(msg)

    def connection_finished(self, exc: Optional[BaseException] = None) -> None:
        super().connection_finished(exc=exc)
        self._stats_logger.connection_finished()

    async def wait_closed(self):
        await self._stats_logger.wait_closed()


def get_logger_receiver() -> Logger:
    return Logger('receiver')


def get_logger_sender() -> Logger:
    return Logger('sender')


def get_connection_logger_receiver() -> ConnectionLogger:
    return ConnectionLogger('receiver.connection')


def get_connection_logger_sender() -> ConnectionLogger:
    return ConnectionLogger('sender.connection')



import asyncio
from pathlib import Path
from ssl import SSLContext

from dataclasses import dataclass, field
import socket
import sys

from aionetworking.compatibility_os import get_client_kwargs
from aionetworking.networking.protocol_factories import DatagramClientProtocolFactory
from aionetworking.types.networking import ConnectionType
from aionetworking.networking.ssl import ClientSideSSL
from .base import BaseClient, BaseNetworkClient

from typing import Union, Optional


@dataclass
class TCPClient(BaseNetworkClient):
    name = "TCP Client"
    peer_prefix = 'tcp'
    transport: asyncio.Transport = field(init=False, compare=False, default=None)
    server_hostname: str = None
    happy_eyeballs_delay: float = None
    interleave: int = None
    ssl: ClientSideSSL = None
    ssl_handshake_timeout: int = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.ssl:
            self.close_tasks.append(self.ssl.close)
            self.ssl.set_logger(self.logger)

    @property
    def ssl_context(self) -> Optional[SSLContext]:
        if self.ssl:
            return self.ssl.context

    async def _open_connection(self) -> ConnectionType:
        extra_kwargs = get_client_kwargs(self.happy_eyeballs_delay, self.interleave)
        self.transport, self.conn = await self.loop.create_connection(
            self.protocol_factory, host=self.host, port=self.port, ssl=self.ssl_context, local_addr=self.local_addr,
            server_hostname=self.server_hostname, ssl_handshake_timeout=self.ssl_handshake_timeout, **extra_kwargs)
        return self.conn


@dataclass
class UnixSocketClient(BaseClient):
    name = "Unix Socket Client"
    peer_prefix = 'unix'
    path: Union[str, Path] = None
    transport: asyncio.Transport = field(init=False, compare=False, default=None)
    server_hostname: Optional[str] = None

    ssl: Optional[ClientSideSSL] = None
    ssl_handshake_timeout: Optional[int] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.ssl:
            self.close_tasks.append(self.ssl.close)
            self.ssl.set_logger(self.logger)

    @property
    def ssl_context(self) -> Optional[SSLContext]:
        if self.ssl:
            return self.ssl.context

    @property
    def src(self) -> str:
        if self.transport:
            _socket = self.transport.get_extra_info('socket')
            return str(_socket.fd)
        return ''

    @property
    def dst(self) -> str:
        return str(self.path)

    async def _open_connection(self) -> ConnectionType:
        extra_kwargs = get_client_kwargs()
        self.transport, self.conn = await self.loop.create_unix_connection(
            self.protocol_factory, path=str(self.path), ssl=self.ssl_context, server_hostname=self.server_hostname,
            ssl_handshake_timeout=self.ssl_handshake_timeout, **extra_kwargs)
        return self.conn



@dataclass
class UDPClient(BaseNetworkClient):
    name = "UDP Client"
    peer_prefix = 'udp'
    transport: asyncio.DatagramTransport = field(init=False, compare=False, default=None)
    reuse_port: Optional[bool] = None
    allow_broadcast: Optional[bool] = None

    async def _open_connection(self) -> ConnectionType:
        protocol_factory: DatagramClientProtocolFactory
        self.transport, protocol_factory = await self.loop.create_datagram_endpoint(
            self.protocol_factory, remote_addr=(self.host, self.port), local_addr=self.local_addr,
            reuse_port=self.reuse_port, allow_broadcast=self.allow_broadcast)
        self.conn = protocol_factory.new_peer()
        await self.conn.wait_connected()
        return self.conn
